"""Module to perform forest inventories using Simple Random Sampling
"""

import math
import logging
import pandas as pd
from scipy import stats


logging.basicConfig(format='%(levelname)s - %(message)s', level=logging.DEBUG)


class RandomSample:
    """Class to create a forest inventory object
    which uses Simple Random Sampling to estimate total wood volume"""

    def __init__(self,
                 dataframe,
                 unit_area,
                 sampling_area,
                 significance=95,
                 sampling_error=20,
                 deg_free=None):

        self.dataframe = dataframe
        self.unit_area = unit_area
        self.sampling_area = sampling_area
        self.significance = significance
        self.sampling_error = sampling_error
        self.deg_free = deg_free
        self.n_sampled_units = self.dataframe.groupby("units").count().shape[0]
        self._add_vol()
        self._get_deg_free()
        self.mean = self.dataframe.groupby("units").sum().mean()[["volume"]]
        self.std = self.dataframe.groupby("units").sum().std()[["volume"]]
        self.var_coef = self.std / self.mean * 100
        self.total_units = self.sampling_area / self.unit_area
        self.t_value = stats.t.ppf(1 - (1 - (self.significance/100)) / 2, self.deg_free)


    def _add_vol(self):

        try:
            self.get_vol()
            logging.warning('Adding colum with Volume.')
        except KeyError:
            if 'volume' in self.dataframe.columns:
                logging.warning('Volume column already exists.')
                return self.dataframe

            logging.warning('Adding columns with DBH and Volume.')
            self.dataframe["dbh"] = self.dataframe["cbh"] / math.pi
            self.get_vol()

        return self.dataframe

    def get_vol(self, beta_0=0.000074, beta_1=1.707348, beta_2=1.16873):
        """Add a column with volume info to the dataframe

        Args:
            beta_0 (float, optional): first parameter. Defaults to 0.000074.
            beta_1 (float, optional): second parameter. Defaults to 1.707348.
            beta_2 (float, optional): third parameter. Defaults to 1.16873.

        Returns:
            dataframe (pd.DataFrame): A DataFrame containing volume info.

        Examples
        --------
        >>> instance.get_vol()
        ### To change the equation used to calculate volume
        >>> instance.get_vol(beta_0=0.000025, beta_1=1.4568, beta_2=1.2563)

        """
        self.dataframe["volume"] = beta_0 * (self.dataframe["dbh"] ** beta_1)\
            * (self.dataframe["height"] ** beta_2)
        return self.dataframe

    def _get_deg_free(self):
        # Get the degrees of freedom
        if self.deg_free is None:
            self.deg_free = self.n_sampled_units - 1
            logging.warning("Degrees of freedom not informed."
                            "Using default value: %s", self.deg_free)
        return self.deg_free

    def get_sample_size(self):
        """Returns the estimated required sample size to perform the
        definitive inventory

        Returns:
            sample size (int): the required sample size

        Examples
        --------
        >>> instance.get_sample_size()
        """
        dividend_sample_size = ((self.t_value ** 2) * (self.var_coef ** 2))
        divisor_sample_size = (self.sampling_error ** 2) + (((self.t_value\
            ** 2) * (self.var_coef ** 2)) / self.total_units)
        sample_size = dividend_sample_size / divisor_sample_size
        return math.ceil(sample_size)

    def srs_inventory(self):
        """Performs the Simple Random Sampling inventory

        Raises:
            ValueError: If the required sample size is greater than the number
            of units sampled, it is not possible to perform the inventory.

        Returns:
            dataframe (pd.DataFrame): A DataFrame containing the inventory
            results.

        Examples
        --------
        >>> instance.srs_inventory()
        """
        if self.get_sample_size() > self.n_sampled_units:
            raise ValueError("Required sample size is greater than "
                             "the number of sampled units "
                             f"{self.get_sample_size()}. "
                             "It is necessary to obtain more data")

        logging.info("Calculating final inventory.")
        var = self.dataframe.groupby("units").sum().var()[["volume"]]

        # standardized error of mean - sem
        sem = math.sqrt(((var / self.n_sampled_units) * (1 -\
            (self.n_sampled_units / self.total_units))))
        total_vol = self.total_units * self.mean
        sample_error = sem * self.t_value
        sample_error_perc = (sample_error / self.mean) * 100

        # CI true mean
        ic_up = self.mean + sample_error
        ic_lo = self.mean - sample_error

        # CI prod by ha
        mean_ha = self.mean * (1 / self.unit_area)
        ic_pop_up = (self.mean + sample_error) * (1 / self.unit_area)
        ic_pop_lo = (self.mean - sample_error) * (1 / self.unit_area)

        # CI vol total pop
        ic_pop_total = self.total_units * self.mean
        ic_pop_up_total = (self.total_units * self.mean) + \
            (self.total_units * sample_error)
        ic_pop_lo_total = (self.total_units * self.mean) - \
            (self.total_units * sample_error)

        t_value_uni = stats.t.ppf(1 - ((1 - (self.significance/100)) * 2) \
            / 2, self.deg_free)

        # Minimum Reliable Estimate - mre
        mre = self.mean - (t_value_uni * sem)
        mre_population = (self.mean - (t_value_uni * sem)) * self.total_units

        results = {"Variáveis": ["Ym", "DesvPad", "CV", "S²", "N parcelas",
                                 "N Total", "ErrMédia", "V Total",
                                "Err", "E(%)", "IC Supe", "IC Infe", "Ym/ha",
                                "IC Supe ha", "IC Infe ha",
                                "Ym/total", "IC Supe total",
                                "IC Infe total", "EMC", "EMC Pop"],
                "Dados": [self.mean.item(), self.std.item(),
                          self.var_coef.item(), var.item(), self.deg_free,
                          self.total_units,
                            sem, total_vol.item(), sample_error,
                            sample_error_perc.item(), ic_up.item(),
                            ic_lo.item(), mean_ha.item(), ic_pop_up.item(),
                            ic_pop_lo.item(), ic_pop_total.item(),
                            ic_pop_up_total.item(), ic_pop_lo_total.item(),
                            mre.item(), mre_population.item()]}

        return pd.DataFrame(data=results)



"""
Module with functions to load the data used in examples.
"""

import pandas as pd


def data_1():
    """
    Loads the data from the board 4.1. - Credits to Soares, Paula Neto & Souza
    (2010).
    The board contains the columns:
        units: the sample units used
        columns: the columns selected from the complete data
        rows: the rows selected from the complete data
        volume: the volume of the sample unit
    Returns:
        data (pandas.DataFrame): The data.

    Examples
    --------
    >>> data_1()

    """
    return pd.read_csv('./src/fivpy/data/data_1.csv', sep=';', decimal=',')


def data_2():
    """
    Loads one of the bult-in datasets with data collected in the Atlantic 
    Forest, Bahia, Brazil.
    The dataset contains the columns:
        n_tree: number of the tree on field
        units: the sample units used
        cbh: circumference at breast height
        height: height of the tree
    Returns:
        data (pandas.DataFrame): The data.

    Examples
    --------
    >>> data_2()

    """
    return pd.read_csv('./src/fivpy/data/data_2.csv', sep=';', decimal=',')

def data_3():
    """
    Loads one of the bult-in datasets with data collected in the Atlantic 
    Forest, Bahia, Brazil.
    The dataset contains the columns:
        n_tree: number of the tree on field
        units: the sample units used
        cbh: circumference at breast height
        height: height of the tree
    Returns:
        data (pandas.DataFrame): The data.

    Examples
    --------
    >>> data_3()

    """
    return pd.read_csv('./src/fivpy/data/data_2.csv', sep=';', decimal=',')



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fivpy', 'fivpy.data']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.3.5,<2.0.0', 'scipy>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'fivpy',
    'version': '0.1.4',
    'description': 'A package for calculating forest inventories in python.',
    'long_description': "# fivpy\n\nA python package to calculate forest inventories.\n\n## Installation\n\n```bash\n$ pip install fivpy\n```\n  \n## Usage\n\n- 'fivpy' can be used to perform forest inventories for a dataset containing the columns units (with the sampled units), dbh or cbh (with the diameter or circunference of   the sampled tress) and height (with the height of the sampled trees).\n\n## Examples\n``` python\nimport pandas as pd\nfrom fivpy import RandomSample\n\n# Loading the data\ndata = pd.read_csv('../src/fivpy/data/data_2.csv', sep=';', decimal=',')\n\n# Creating a inventory instance\ninventory_1 = RandomSample(data,\n                            unit_area=0.02,\n                            sampling_area=11,\n                            significance=90,\n                            sampling_error=10)\n\n# Calculating inventory atributes\ninventory_1.srs_inventory()\n```\nFor more infos and examples of usage please visit **https://fivpy.readthedocs.io/**\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`fivpy` was created by Theilon Macedo. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`fivpy` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).  \n\nThe metodology used here is based on the book 'Dendrometria e Inventário Florestal'. **SOARES, C.P.B.; PAULA NETO. F.; SOUZA, A.L. Dendrometria e inventário florestal. 2. Ed. Viçosa: Editora UFV. 2011. 272 p.**",
    'author': 'Theilon Macedo',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



# read version from installed package
from importlib.metadata import version
__version__ = version("fivpy")


from fivpy.srs import RandomSample
from fivpy.datasets import data_1
from fivpy.datasets import data_2
from fivpy.datasets import data_3